#ifndef ENGINE_H
#define ENGINE_H

#include <vector>
#include <list>
#include <allegro.h>
#include <fstream>
#include "map.h"
#include <map>

using namespace std;

const int N = 0;
const int E = 1;
const int S = 2;
const int W = 3;

class Train
{
	public:
	static int getDx(int d)
	{
		if (d == E) return 1;
		if (d == W) return -1;
		return 0;
	}
	
	static int getDy(int d)
	{
		if (d == N) return -1;
		if (d == S) return 1;
		return 0;
	}
	
	float getLx()
	{		
		return last_node_x + steps * (float)getDx(dir);
	}

	float getLy()
	{		
		return last_node_y + steps * (float)getDy(dir);
	}
	
	int getNextNodeX()
	{
		return last_node_x + getDx(dir);
	}

	int getNextNodeY()
	{
		return last_node_y + getDy(dir);
	}
	
	int getCurrentGoal()
	{
		assert (currentStep >= 0 && currentStep < (int)program.size());
		return program[currentStep];		
	}
		
	bool serialize (fstream &f) const;
	bool deserialize (fstream &f) const;
	
	Train();
	void doMove();

	bool isDead; // waiting to be re-used
	
	vector <int> program; // list of station id's
	int currentStep; // current step in program
	
	int id;
	int dir;
	
	float speed;
	float acceleration;
	float max_speed;
	
	int last_node_x; // last visited node
	int last_node_y;
	
	
	bool hasPassengers; 
	int idCityPassengersSrc;
	int passengersDepartureTick;
	
	float steps; // after steps > 1.0, we arrive at the next node
	
	list <int> path;	
};

//~ class Passenger
//~ {
	//~ public:
	//~ bool serialize (fstream &f) const;
	//~ bool deserialize (fstream &f) const;
	
	//~ int number; // number of people in this passenger group
	//~ int currentStation; // current station, -1 if on train
	//~ int currentTrain; // current train, -1 if on station
	//~ int srcCity; // source city
	//~ int destCity; // dest city
	//~ int startTick; // when started traveling?
//~ };

//~ class Arrival
//~ {
	//~ public:
	//~ bool serialize (fstream &f) const;
	//~ bool deserialize (fstream &f) const;

	//~ int tick; // arrival time
	//~ int duration; // amount of time from start to arrival
	//~ int number; // number of passengers
//~ };

// stores station x station cross-info

class Station
{
	public:
	bool serialize (fstream &f) const;
	bool deserialize (fstream &f) const;
	Station();
	
	bool isDead; // waiting to be re-used
	int id;
	
	int idCity; // nearest city
	string name;
	int x, y; // topleft cell
	int w, h; // space occupation
};

class City
{
	public:
	City();
	bool serialize (fstream &f) const;
	bool deserialize (fstream &f) const;
	
	bool isDead; // waiting to be re-used
	int id;
	
	string name;
	int population;
	int center_x;
	int center_y;
	
	map <int, int> timeTable;
};

class Render
{
	public:
	void init();
	void draw (BITMAP *bmp, int xofst, int yofst);
	int getw() const;
	int geth() const;
	int canvasFromMapY (float mx, float my) const;
	int canvasFromMapX (float mx, float my) const;
	int canvasFromMapY (int mx, int my) const;
	int canvasFromMapX (int mx, int my) const;
	int mapFromCanvasX (int x, int y) const;
	int mapFromCanvasY (int x, int y) const;
	
	private:
		
	RLE_SPRITE *tracks[18];
	RLE_SPRITE *buildings[9];
	RLE_SPRITE *wagon[2];
};

extern Render render;

const int UNDEFINED = -1;
const int OPEN = 1;
const int CLOSED = 2;

class AstarTile
{
	public:
	float h;
	float g;
	int state; // -1 undef, 1 open, 2 closed
	int parent_x;
	int parent_y;
};

class AstarNode
{
	public:
	float f;
	int mx;
	int my;
};

// astar node comparator

class Engine
{
	public:
	
	// public only for debugging...
	AstarTile *temp_map;
	vector <AstarNode> nodes;
	
	int minFromTick (int t) { return tick / 100; }
	int getMin() { return minFromTick (tick); }
	int connectionsLeft;

	bool displayLabels;
	bool fullRedraw;
	
	private:
		
	void examineCell (int idTrain, int x, int y, int d);
	float getHeuristics (int sx, int sy, int dx, int dy);	
	void openCell (int mx, int my, int px, int py, float newg, float newh);
	void getBestOpenCell (int &x, int &y);
	void generateMap();

	void setArrivalTime (int idSrcCity, int idDestCity, int dTime);

	public:

	int getArrivalTime (int idSrcCity, int idDestCity); // returns -1 if no arrival time

	Engine();
	~Engine();
	
	bool isGoal (int idTrain, int mx, int my);
	bool canGo (int mx, int my, int d);
	void createPath (int idTrain);
	void init();
	void initGame();
	void update();
	int addTrain(int idStation); // returns -1 on failure
	int addStation (int mx, int my); // returns -1 on failure
	int addCity(int mx, int my, string name);
	void onArrive (int idTrain, int idStation);
	
	bool serialize (fstream &f) const;
	bool deserialize (fstream &f) const;

	int tick;
	int money;
	Map map;
	
	
	vector <City> city;
	vector <Train> train;
	vector <Station> station;
	//~ list <Passenger> passenger;	
	
};

extern Engine engine;

#endif
